@@ -12,6 +12,7 @@
 #include <unistd.h>
 #include <syslog.h>
 #include <stdlib.h>
+#include <grp.h>
 #include <libgen.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -23,6 +24,7 @@
 #include <stdarg.h>
 #include <getopt.h>
 #include <dlfcn.h>
+#include <limits.h>
 
 #include <config.h>
 #include <vde.h>
@@ -48,6 +50,8 @@ static unsigned int console_type=-1;
 static unsigned int mgmt_ctl=-1;
 static unsigned int mgmt_data=-1;
 static int mgmt_mode = 0600;
+static gid_t mgmt_group = -1;
+
 static char *mgmt_socket = NULL;
 static char header[]="VDE switch V.%s\n(C) Virtual Square Team (coord. R. Davoli) 2005,2006,2007 - GPLv2\n";
 static char prompt[]="\nvde$ ";
@@ -164,27 +168,6 @@ void printlog(int priority, const char *format, ...)
 	va_end (arg);
 }
 
-#if 0
-void printoutc(int fd, const char *format, ...)
-{
-	va_list arg;
-
-	va_start (arg, format);
-#if 0
-	if (fd < 0)
-		printlog(LOG_INFO,format,arg);
-	else {
-#endif
-		char outbuf[MAXCMD+1];
-		vsnprintf(outbuf,MAXCMD,format,arg);
-		strcat(outbuf,"\n");
-		write(fd,outbuf,strlen(outbuf));
-#if 0
-	}
-#endif
-}
-#endif
-
 void printoutc(FILE *f, const char *format, ...)
 {
 	va_list arg;
@@ -227,7 +210,7 @@ void eventout(struct dbgcl* cl, ...)
 	va_list arg;
 	for (i=0; i<cl->nfun; i++) {
 		va_start (arg, cl);
-		(cl->fun[i])(cl,arg);
+		(cl->fun[i])(cl,cl->funarg[i],arg);
 		va_end(arg);
 	}
 }
@@ -244,7 +227,7 @@ int packetfilter(struct dbgcl* cl, ...)
 	va_end(arg);
 	for (i=0; i<cl->nfun && len>0; i++) {
 		va_start (arg, cl);
-		int rv=(cl->fun[i])(cl,arg);
+		int rv=(cl->fun[i])(cl,cl->funarg[i],arg);
 		va_end (arg);
 		if (rv!=0) 
 			len=rv;
@@ -259,6 +242,7 @@ int packetfilter(struct dbgcl* cl, ...)
 void setmgmtperm(char *path)
 {
 	chmod(path,mgmt_mode);
+	chown(path, -1, mgmt_group);
 }
 
 static int help(FILE *fd,char *arg)
@@ -281,11 +265,7 @@ static int handle_cmd(int type,int fd,char *inbuf)
 	if (*inbuf != '\0' && *inbuf != '#') {
 		char *outbuf;
 		size_t outbufsize;
-		FILE *f;
-		if (fd >= 0)
-			f=open_memstream(&outbuf,&outbufsize);
-		else
-			f=NULL;
+		FILE *f=open_memstream(&outbuf,&outbufsize);
 		for (p=clh;p!=NULL && (p->doit==NULL || strncmp(p->path,inbuf,strlen(p->path))!=0); p=p->next)
 			;
 		if (p!=NULL)
@@ -293,21 +273,24 @@ static int handle_cmd(int type,int fd,char *inbuf)
 			inbuf += strlen(p->path);
 			while (*inbuf == ' ' || *inbuf == '\t') inbuf++;
 			if (p->type & WITHFD) {
-				if (p->type & WITHFILE) {
-					printoutc(f,"0000 DATA END WITH '.'");
-					switch(p->type & ~(WITHFILE | WITHFD)){
-						case NOARG: rv=p->doit(f,fd); break;
-						case INTARG: rv=p->doit(f,fd,atoi(inbuf)); break;
-						case STRARG: rv=p->doit(f,fd,inbuf); break;
+				if (fd >= 0) {
+					if (p->type & WITHFILE) {
+						printoutc(f,"0000 DATA END WITH '.'");
+						switch(p->type & ~(WITHFILE | WITHFD)){
+							case NOARG: rv=p->doit(f,fd); break;
+							case INTARG: rv=p->doit(f,fd,atoi(inbuf)); break;
+							case STRARG: rv=p->doit(f,fd,inbuf); break;
+						}
+						printoutc(f,".");
+					} else {
+						switch(p->type & ~WITHFD){
+							case NOARG: rv=p->doit(fd); break;
+							case INTARG: rv=p->doit(fd,atoi(inbuf)); break;
+							case STRARG: rv=p->doit(fd,inbuf); break;
+						}
 					}
-					printoutc(f,".");
-				} else {
-					switch(p->type & ~WITHFD){
-						case NOARG: rv=p->doit(fd); break;
-						case INTARG: rv=p->doit(fd,atoi(inbuf)); break;
-						case STRARG: rv=p->doit(fd,inbuf); break;
-					}
-				}
+				} else
+					rv = EBADF;
 			} else if (p->type & WITHFILE) {
 				printoutc(f,"0000 DATA END WITH '.'");
 				switch(p->type & ~WITHFILE){
@@ -324,13 +307,12 @@ static int handle_cmd(int type,int fd,char *inbuf)
 				}
 			}
 		}
-		if (rv >= 0 && (rv > 0 || fd >= 0))
+		if (rv >=0)
 			printoutc(f,"1%03d %s",rv,strerror(rv));
-		if (f) {
-			fclose(f);
+		fclose(f);
+		if (fd >= 0)
 			write(fd,outbuf,outbufsize);
-			free(outbuf);
-		}
+		free(outbuf);
 	}
 	return rv;
 }
@@ -346,7 +328,7 @@ static int runscript(int fd,char *path)
 			if (strlen(buf) > 1 && buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]= '\0';
 			if (fd >= 0) {
 				char *scriptprompt=NULL;
-				asprintf(&scriptprompt,"vde[%s]: %s",path,buf);
+				asprintf(&scriptprompt,"vde[%s]: %s\n",path,buf);
 				write(fd,scriptprompt,strlen(scriptprompt));
 				free(scriptprompt);
 			}
@@ -505,6 +487,7 @@ static void cleanup(unsigned char type,int fd,int arg)
 }
 
 #define MGMTMODEARG 0x100
+#define MGMTGROUPARG 0x101
 
 static struct option long_options[] = {
 	{"daemon", 0, 0, 'd'},
@@ -512,6 +495,7 @@ static struct option long_options[] = {
 	{"rcfile", 1, 0, 'f'},
 	{"mgmt", 1, 0, 'M'},
 	{"mgmtmode", 1, 0, MGMTMODEARG},
+	{"mgmtgroup", 1, 0, MGMTGROUPARG},
 #ifdef DEBUGOPT
 	{"debugclients",1,0,'D'},
 #endif
@@ -528,8 +512,9 @@ static void usage(void)
 			"  -f, --rcfile               Configuration file (overrides %s and ~/.vderc)\n"
 			"  -M, --mgmt SOCK            path of the management UNIX socket\n"
 			"      --mgmtmode MODE        management UNIX socket access mode (octal)\n"
+			"      --mgmtgroup GROUP      management UNIX socket group name\n"
 #ifdef DEBUGOPT
-			"  -D, --debugclients #        number of debug clients allowed\n"
+			"  -D, --debugclients #       number of debug clients allowed\n"
 #endif
 			,STDRCFILE);
 }
@@ -537,6 +522,7 @@ static void usage(void)
 static int parseopt(int c, char *optarg)
 {
 	int outc=0;
+	struct group *grp;
 	switch (c) {
 		case 'd':
 			daemonize=1;
@@ -553,6 +539,15 @@ static int parseopt(int c, char *optarg)
 		case MGMTMODEARG:
 			sscanf(optarg,"%o",&mgmt_mode);
 			break;
+		case MGMTGROUPARG:
+			if (!(grp = getgrnam(optarg)))
+			{
+				fprintf(stderr, "No such group '%s'\n", optarg);
+				exit(1);
+			}
+			mgmt_group = grp->gr_gid;
+			break;
+
 		default:
 			outc=c;
 	}
@@ -617,7 +612,7 @@ static void init(void)
 				return;
 			}
 		}
-		chmod(sun.sun_path,mgmt_mode);
+		setmgmtperm(sun.sun_path);
 		if(listen(mgmtconnfd, 15) < 0){
 			printlog(LOG_ERR,"mgmt listen: %s",strerror(errno));
 			return;
@@ -825,19 +820,21 @@ static int pluginadd(char *arg) {
 }
 
 static int plugindel(char *arg) {
-	int rv=ENOENT;
 	struct plugin **p=&pluginh;
 	while (*p!=NULL){
+		void *handle;
 		if (strncmp((*p)->name, arg, strlen(arg)) == 0
 				&& ((*p)->handle != NULL)) {
 			struct plugin *this=*p;
 			delplugin(this);
-			dlclose(this->handle);
-			rv=0;
+			handle=this->handle;
+			this->handle=NULL;
+			dlclose(handle);
+			return 0;
 		} else
 			p=&(*p)->next;
 	}
-	return rv;
+	return ENOENT;
 }
 #endif
 
