@@ -13,6 +13,10 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h> /*ntoh conversion*/
+#include <sys/types.h>
+#include <grp.h>
+#include <pwd.h>
+#include <ctype.h>
 
 #include <config.h>
 #include <vde.h>
@@ -87,10 +91,13 @@ struct port {
 	int fd_data;
 	struct endpoint *ep;
 	int flag;
-	/* sender is already inside ep, but it needs one more memaccess */
+	/* sender is already inside ms, but it needs one more memaccess */
 	int (*sender)(int fd, int fd_ctl, void *packet, int len, void *data, int port);
 	struct mod_support *ms;
 	int vlanuntag;
+	uid_t user;
+	gid_t group;
+	uid_t curuser;
 #ifdef FSTP
 	int cost;
 #endif
@@ -143,6 +150,7 @@ static int alloc_port(unsigned int portno)
 				portv[i]=port;
 				port->fd_data=-1;
 				port->ep=NULL;
+				port->user=port->group=port->curuser=-1;
 #ifdef FSTP
 				port->cost=DEFAULT_COST;
 #endif
@@ -155,7 +163,7 @@ static int alloc_port(unsigned int portno)
 				port->flag=0;
 				port->sender=NULL;
 				port->vlanuntag=0;
-				BA_SET(vlant[0].table,i);
+				ba_set(vlant[0].table,i);
 			}
 		}
 		return i;
@@ -170,17 +178,90 @@ static void free_port(unsigned int portno)
 			portv[portno]=NULL;
 			register int i;
 			/* delete completely the port. all vlan defs zapped */
-			BAC_FORALL(validvlan,NUMOFVLAN,BA_CLR(vlant[i].table,portno),i);
+			bac_FORALL(validvlan,NUMOFVLAN,ba_clr(vlant[i].table,portno),i);
 			free(port);
 		}
 	}
 }
 
+/* 1 if user belongs to the group, 0 otherwise) */
+#if 0
+/* getgrouplist is a nonstandard call */
+static int user_belongs_to_group(uid_t uid, gid_t gid)
+{
+	struct passwd *pw=getpwuid(uid);
+	int found=0;
+	if (pw != NULL) {
+		int gsize=8;
+		int scan;
+		gid_t *grouplist=NULL;
+		do {
+			grouplist=realloc(grouplist, gsize*sizeof(gid_t));
+		} while (grouplist != NULL && getgrouplist(pw->pw_name,pw->pw_gid,grouplist,&gsize) < 0);
+		if (grouplist != NULL) {
+			for (scan=0;scan<gsize && !found;scan++)
+				if (grouplist[scan]==gid)
+					found=1;
+			free(grouplist);
+		}
+	}
+	return found;
+}
+#endif
+static int user_belongs_to_group(uid_t uid, gid_t gid)
+{
+	struct passwd *pw=getpwuid(uid);
+	if (pw == NULL) 
+		return 0;
+	else {
+		if (gid==pw->pw_gid)
+			return 1;
+		else {
+			struct group *grp;
+			setgrent();
+			while ((grp = getgrent())) {
+				if (grp->gr_gid == gid) {
+					int i;
+					for (i = 0; grp->gr_mem[i]; i++) {
+						if (strcmp(grp->gr_mem[i], pw->pw_name)==0) {
+							endgrent();
+							return 1;
+						}
+					}
+				}
+			}
+			endgrent();
+			return 0;
+		}
+	}
+}
+
+
+/* Access Control check:
+	 returns 0->OK -1->Permission Denied */
+static int checkport_ac(struct port *port, uid_t user)
+{
+	/*unrestricted*/
+	if (port->user == -1 && port->group == -1)
+		return 0;
+	/*root or restricted to a specific user*/
+	else if (user==0 || (port->user != -1 && port->user==user))
+		return 0;
+	/*restricted to a group*/
+	else if (port->group != -1 && user_belongs_to_group(user,port->group))
+		return 0;
+	else {
+		errno=EPERM;
+		return -1;
+	}
+}
+
 /* initialize a port structure with control=fd, given data+data_len and sender
  * function; 
  * and then add it to the g_fdsdata array at index i. */
 int setup_ep(int portno, int fd_ctl,
 		void *data,
+		uid_t user,
 		struct mod_support *modfun)
 {
 	struct port *port;
@@ -188,9 +269,11 @@ int setup_ep(int portno, int fd_ctl,
 
 	if ((portno = alloc_port(portno)) >= 0) {
 		port=portv[portno];	
-		if (port->fd_data < 0)
-			port->fd_data=modfun->newport(fd_ctl,portno);
-		if (port->fd_data >= 0 &&
+		if (port->fd_data < 0 && checkport_ac(port,user)==0) {
+			port->fd_data=modfun->newport(fd_ctl,portno,user);
+			port->curuser=user;
+		}
+		if (port->fd_data >= 0 && port->curuser == user &&
 				(ep=malloc(sizeof(struct endpoint))) != NULL) {
 			DBGOUT(DBGEPNEW,"Port %02d FD %2d", portno,fd_ctl);
 			EVENTOUT(DBGEPNEW,portno,fd_ctl);
@@ -205,19 +288,19 @@ int setup_ep(int portno, int fd_ctl,
 				/* copy all the vlan defs to the active vlan defs */
 				ep->next=port->ep;
 				port->ep=ep;
-				BAC_FORALL(validvlan,NUMOFVLAN,
-						({if (BA_CHECK(vlant[i].table,portno)) {
-						 BA_SET(vlant[i].bctag,portno);
+				bac_FORALL(validvlan,NUMOFVLAN,
+						({if (ba_check(vlant[i].table,portno)) {
+						 ba_set(vlant[i].bctag,portno);
 #ifdef FSTP
 						 fstaddport(i,portno,(i!=port->vlanuntag));
 #endif
 						 }
 						 }),i);
 				if (port->vlanuntag != NOVLAN) {
-					BA_SET(vlant[port->vlanuntag].bcuntag,portno);
-					BA_CLR(vlant[port->vlanuntag].bctag,portno);
+					ba_set(vlant[port->vlanuntag].bcuntag,portno);
+					ba_clr(vlant[port->vlanuntag].bctag,portno);
 				}
-				BA_CLR(vlant[port->vlanuntag].notlearning,portno);
+				ba_clr(vlant[port->vlanuntag].notlearning,portno);
 			} else {
 				ep->next=port->ep;
 				port->ep=ep;
@@ -225,11 +308,19 @@ int setup_ep(int portno, int fd_ctl,
 			return portno;
 		}
 		else {
+			if (port->fd_data < 0)
+				errno=EPERM;
+			else if (port->curuser != user)
+				errno=EADDRINUSE;
+			else 
+				errno=ENOMEM;
 			return -1;
 		}
 	}
-	else
+	else {
+		errno=ENOMEM;
 		return -1;
+	}
 }
 
 void setup_description(int portno, int fd_ctl, char *descr)
@@ -284,21 +375,22 @@ int close_ep(int portno, int fd_ctl)
 				port->fd_data=-1;
 				port->ms=NULL;
 				port->sender=NULL;
+				port->curuser=-1;
 				register int i;
 				/* inactivate port: all active vlan defs cleared */
-				BAC_FORALL(validvlan,NUMOFVLAN,({
-							BA_CLR(vlant[i].bctag,portno);
+				bac_FORALL(validvlan,NUMOFVLAN,({
+							ba_clr(vlant[i].bctag,portno);
 #ifdef FSTP
 							fstdelport(i,portno);
 #endif
 							}),i);
-				if (port->vlanuntag < NOVLAN) BA_CLR(vlant[port->vlanuntag].bcuntag,portno);
+				if (port->vlanuntag < NOVLAN) ba_clr(vlant[port->vlanuntag].bcuntag,portno);
 			}
 			return rv;	
 		} else
-		return ENXIO;
+			return ENXIO;
 	} else
-	return EINVAL;
+		return EINVAL;
 }
 
 int portflag(int op,int f)
@@ -325,19 +417,19 @@ int portflag(int op,int f)
 
 #ifdef VDE_PQ
 #define SEND_PACKET_PORT(PORT,PORTNO,PACKET,LEN) \
-	 ({\
+	({\
 	 struct port *Port=(PORT); \
 	 if (PACKETFILTER(PKTFILTOUT,(PORTNO),(PACKET), (LEN))) {\
 	 struct endpoint *ep; \
 	 SEND_COUNTER_UPD(Port,LEN); \
 	 for (ep=Port->ep; ep != NULL; ep=ep->next) \
 	 if (Port->ms->sender(Port->fd_data, ep->fd_ctl, (PACKET), (LEN), ep->data, ep->port)) \
-	 	packetq_add(Port->ms->sender,Port->fd_data, ep->fd_ctl, (PACKET), (LEN), ep->data, ep->port); \
+	 packetq_add(Port->ms->sender,Port->fd_data, ep->fd_ctl, (PACKET), (LEN), ep->data, ep->port); \
 	 } \
-		})
+	 })
 #else
 #define SEND_PACKET_PORT(PORT,PORTNO,PACKET,LEN) \
-	 ({\
+	({\
 	 struct port *Port=(PORT); \
 	 if (PACKETFILTER(PKTFILTOUT,(PORTNO),(PACKET), (LEN))) {\
 	 struct endpoint *ep; \
@@ -345,7 +437,7 @@ int portflag(int op,int f)
 	 for (ep=Port->ep; ep != NULL; ep=ep->next) \
 	 Port->ms->sender(Port->fd_data, ep->fd_ctl, (PACKET), (LEN), ep->data, ep->port); \
 	 } \
-		})
+	 })
 #endif
 
 #ifdef FSTP
@@ -359,39 +451,39 @@ void port_send_packet(int portno, void *packet, int len)
 void portset_send_packet(bitarray portset, void *packet, int len)
 {
 	register int i;
-	BA_FORALL(portset,numports,
+	ba_FORALL(portset,numports,
 			SEND_PACKET_PORT(portv[i],i,packet,len), i);
 }
 
 
 void port_set_status(int portno, int vlan, int status)
 {
-	if (BA_CHECK(vlant[vlan].table,portno)) {
+	if (ba_check(vlant[vlan].table,portno)) {
 		if (status==DISCARDING) {
-			BA_SET(vlant[vlan].notlearning,portno);
-			BA_CLR(vlant[vlan].bctag,portno);
-			BA_CLR(vlant[vlan].bcuntag,portno);
+			ba_set(vlant[vlan].notlearning,portno);
+			ba_clr(vlant[vlan].bctag,portno);
+			ba_clr(vlant[vlan].bcuntag,portno);
 		} else if (status==LEARNING) {
-			BA_CLR(vlant[vlan].notlearning,portno);
-			BA_CLR(vlant[vlan].bctag,portno);
-			BA_CLR(vlant[vlan].bcuntag,portno);
+			ba_clr(vlant[vlan].notlearning,portno);
+			ba_clr(vlant[vlan].bctag,portno);
+			ba_clr(vlant[vlan].bcuntag,portno);
 		} else { /*forwarding*/
-			BA_CLR(vlant[vlan].notlearning,portno);
+			ba_clr(vlant[vlan].notlearning,portno);
 			if (portv[portno]->vlanuntag == vlan) 
-				BA_SET(vlant[vlan].bcuntag,portno);
+				ba_set(vlant[vlan].bcuntag,portno);
 			else 
-				BA_SET(vlant[vlan].bctag,portno);
+				ba_set(vlant[vlan].bctag,portno);
 		}
 	}
 }
 
 int port_get_status(int portno, int vlan)
 {
-	if (BA_CHECK(vlant[vlan].notlearning,portno)) 
+	if (ba_check(vlant[vlan].notlearning,portno)) 
 		return DISCARDING;
 	else {
-		if (BA_CHECK(vlant[vlan].bctag,portno) ||
-				BA_CHECK(vlant[vlan].bcuntag,portno))
+		if (ba_check(vlant[vlan].bctag,portno) ||
+				ba_check(vlant[vlan].bcuntag,portno))
 			return FORWARDING;
 		else
 			return LEARNING;
@@ -466,7 +558,7 @@ void handle_in_packet(int port,  struct packet *packet, int len)
 			if (packet->header.proto[0] == 0x81 && packet->header.proto[1] == 0x00) {
 				tagged=1;
 				vlan=((packet->data[0] << 8) + packet->data[1]) & 0xfff;
-				if (! BA_CHECK(vlant[vlan].table,port))
+				if (! ba_check(vlant[vlan].table,port))
 					return; /*discard unwanted packets*/
 			} else {
 				tagged=0;
@@ -482,7 +574,7 @@ void handle_in_packet(int port,  struct packet *packet, int len)
 			}
 #endif
 			/* The port is in blocked status, no packet received */
-			if (BA_CHECK(vlant[vlan].notlearning,port)) return; 
+			if (ba_check(vlant[vlan].notlearning,port)) return; 
 
 			/* We don't like broadcast source addresses */
 			if(! (IS_BROADCAST(packet->header.src))) {
@@ -502,17 +594,17 @@ void handle_in_packet(int port,  struct packet *packet, int len)
 				 * of the same tag-ness, then transform it to the other tag-ness for the others*/
 				if (tagged) {
 					register int i;
-					BA_FORALL(vlant[vlan].bctag,numports,
+					ba_FORALL(vlant[vlan].bctag,numports,
 							({if (i != port) SEND_PACKET_PORT(portv[i],i,packet,len);}),i);
 					packet=TAG2UNTAG(packet,len);
-					BA_FORALL(vlant[vlan].bcuntag,numports,
+					ba_FORALL(vlant[vlan].bcuntag,numports,
 							({if (i != port) SEND_PACKET_PORT(portv[i],i,packet,len);}),i);
 				} else { /* untagged */
 					register int i;
-					BA_FORALL(vlant[vlan].bcuntag,numports,
+					ba_FORALL(vlant[vlan].bcuntag,numports,
 							({if (i != port) SEND_PACKET_PORT(portv[i],i,packet,len);}),i);
 					packet=UNTAG2TAG(packet,vlan,len);
-					BA_FORALL(vlant[vlan].bctag,numports,
+					ba_FORALL(vlant[vlan].bctag,numports,
 							({if (i != port) SEND_PACKET_PORT(portv[i],i,packet,len);}),i);
 				}
 			}
@@ -566,21 +658,28 @@ static int portsetnumports(int val)
 		}
 		for (i=0;i<NUMOFVLAN;i++) { 
 			if (vlant[i].table) {
-				vlant[i].table=BA_REALLOC(vlant[i].table,numports,val);
+				vlant[i].table=ba_realloc(vlant[i].table,numports,val);
 				if (vlant[i].table == NULL) {
 					printlog(LOG_ERR,"Numport resize failed vlan tables vlan table %s",strerror(errno));
 					exit(1);
 				}
 			}
 			if (vlant[i].bctag) {
-				vlant[i].bctag=BA_REALLOC(vlant[i].bctag,numports,val);
+				vlant[i].bctag=ba_realloc(vlant[i].bctag,numports,val);
 				if (vlant[i].bctag == NULL) {
 					printlog(LOG_ERR,"Numport resize failed vlan tables vlan bctag %s",strerror(errno));
 					exit(1);
 				}
 			}
+			if (vlant[i].bcuntag) {
+				vlant[i].bcuntag=ba_realloc(vlant[i].bcuntag,numports,val);
+				if (vlant[i].bcuntag == NULL) {
+					printlog(LOG_ERR,"Numport resize failed vlan tables vlan bctag %s",strerror(errno));
+					exit(1);
+				}
+			}
 			if (vlant[i].notlearning) {
-				vlant[i].notlearning=BA_REALLOC(vlant[i].notlearning,numports,val);
+				vlant[i].notlearning=ba_realloc(vlant[i].notlearning,numports,val);
 				if (vlant[i].notlearning == NULL) {
 					printlog(LOG_ERR,"Numport resize failed vlan tables vlan notlearning %s",strerror(errno));
 					exit(1);
@@ -614,6 +713,56 @@ static int portallocatable(char *arg)
 	return 0;
 }
 
+static int portsetuser(char *arg)
+{
+	int port;
+	char *portuid=arg;
+	struct passwd *pw;
+	while (*portuid != 0 && *portuid == ' ') portuid++;
+	while (*portuid != 0 && *portuid != ' ') portuid++;
+	while (*portuid != 0 && *portuid == ' ') portuid++;
+	if (sscanf(arg,"%i",&port) != 1 || *portuid==0)
+		return EINVAL;
+	if (port < 0 || port >= numports)
+		return EINVAL;
+	if (portv[port] == NULL)
+		return ENXIO;
+	if ((pw=getpwnam(portuid)) != NULL)
+		portv[port]->user=pw->pw_uid;
+	else if (isdigit(*portuid)) 
+		portv[port]->user=atoi(portuid);
+	else if (strcmp(portuid,"NONE")==0 || strcmp(portuid,"ANY")==0) 
+		portv[port]->user= -1;
+	else
+		return EINVAL;
+	return 0;
+}
+
+static int portsetgroup(char *arg)
+{
+	int port;
+	char *portgid=arg;
+	struct group *gr;
+	while (*portgid != 0 && *portgid == ' ') portgid++;
+	while (*portgid != 0 && *portgid != ' ') portgid++;
+	while (*portgid != 0 && *portgid == ' ') portgid++;
+	if (sscanf(arg,"%i",&port) != 1 || *portgid==0)
+		return EINVAL;
+	if (port < 0 || port >= numports)
+		return EINVAL;
+	if (portv[port] == NULL)
+		return ENXIO;
+	if ((gr=getgrnam(portgid)) != NULL)
+		portv[port]->group=gr->gr_gid;
+	else if (isdigit(*portgid)) 
+		portv[port]->group=atoi(portgid);
+	else if (strcmp(portgid,"NONE")==0 || strcmp(portgid,"ANY")==0) 
+		portv[port]->group= -1;
+	else
+		return EINVAL;
+	return 0;
+}
+
 static int portremove(int val)
 {
 	if (val <0 || val>=numports)
@@ -647,6 +796,40 @@ static int epclose(char *arg)
 		return close_ep(port,id);
 }
 
+static char *port_getuser(uid_t uid)
+{
+	static char buf[6];
+	struct passwd *pw;
+	if (uid == -1) 
+		return "NONE";
+	else {
+		pw=getpwuid(uid);
+		if (pw != NULL)
+			return pw->pw_name;
+		else {
+			sprintf(buf,"%d",uid);
+			return buf;
+		}
+	}
+}
+
+static char *port_getgroup(gid_t gid)
+{
+	static char buf[6];
+	struct group *gr;
+	if (gid == -1) 
+		return "NONE";
+	else {
+		gr=getgrgid(gid);
+		if (gr != NULL)
+			return gr->gr_name;
+		else {
+			sprintf(buf,"%d",gid);
+			return buf;
+		}
+	}
+}
+
 static int print_port(FILE *fd,int i,int inclinactive)
 {
 	struct endpoint *ep;
@@ -655,6 +838,10 @@ static int print_port(FILE *fd,int i,int inclinactive)
 				i,portv[i]->vlanuntag,
 				portv[i]->ep?"":"IN",
 				(portv[i]->flag & NOTINPOOL)?"NOT ":"");
+		printoutc(fd," Current User: %s Access Control: (User: %s - Group: %s)", 
+				port_getuser(portv[i]->curuser),
+				port_getuser(portv[i]->user), 
+				port_getgroup(portv[i]->group));
 #ifdef PORTCOUNTERS
 		printoutc(fd," IN:  pkts %10lld          bytes %20lld",portv[i]->pktsin,portv[i]->bytesin);
 		printoutc(fd," OUT: pkts %10lld          bytes %20lld",portv[i]->pktsout,portv[i]->bytesout);
@@ -751,7 +938,7 @@ static int portsetvlan(char *arg)
 	/* port NOVLAN is okay here, it means NO untagged traffic */
 	if (vlan <0 || vlan > NUMOFVLAN || port < 0 || port >= numports) 
 		return EINVAL;
-	if ((vlan != NOVLAN && !BAC_CHECK(validvlan,vlan)) || portv[port] == NULL)
+	if ((vlan != NOVLAN && !bac_check(validvlan,vlan)) || portv[port] == NULL)
 		return ENXIO;
 	int oldvlan=portv[port]->vlanuntag;
 	portv[port]->vlanuntag=NOVLAN;
@@ -759,18 +946,18 @@ static int portsetvlan(char *arg)
 	if (portv[port]->ep != NULL) {
 		/*changing active port*/
 		if (oldvlan != NOVLAN) 
-			BA_CLR(vlant[oldvlan].bcuntag,port);
+			ba_clr(vlant[oldvlan].bcuntag,port);
 		if (vlan != NOVLAN) {
-			BA_SET(vlant[vlan].bcuntag,port);
-			BA_CLR(vlant[vlan].bctag,port);
+			ba_set(vlant[vlan].bcuntag,port);
+			ba_clr(vlant[vlan].bctag,port);
 		}
 #ifdef FSTP
 		if (oldvlan != NOVLAN) fstdelport(oldvlan,port);
 		if (vlan != NOVLAN) fstaddport(vlan,port,0);
 #endif
 	}
-	if (oldvlan != NOVLAN) BA_CLR(vlant[oldvlan].table,port);
-	if (vlan != NOVLAN) BA_SET(vlant[vlan].table,port);
+	if (oldvlan != NOVLAN) ba_clr(vlant[oldvlan].table,port);
+	if (vlan != NOVLAN) ba_set(vlant[vlan].table,port);
 	portv[port]->vlanuntag=vlan;
 	return 0;
 }
@@ -778,10 +965,10 @@ static int portsetvlan(char *arg)
 static int vlancreate_nocheck(int vlan)
 {
 	int rv=0;
-	vlant[vlan].table=BA_ALLOC(numports);
-	vlant[vlan].bctag=BA_ALLOC(numports);
-	vlant[vlan].bcuntag=BA_ALLOC(numports);
-	vlant[vlan].notlearning=BA_ALLOC(numports);
+	vlant[vlan].table=ba_alloc(numports);
+	vlant[vlan].bctag=ba_alloc(numports);
+	vlant[vlan].bcuntag=ba_alloc(numports);
+	vlant[vlan].notlearning=ba_alloc(numports);
 	if (vlant[vlan].table == NULL || vlant[vlan].bctag == NULL || 
 			vlant[vlan].bcuntag == NULL) 
 		return ENOMEM;
@@ -790,7 +977,7 @@ static int vlancreate_nocheck(int vlan)
 		rv=fstnewvlan(vlan);
 #endif
 		if (rv == 0) {
-			BAC_SET(validvlan,NUMOFVLAN,vlan);
+			bac_set(validvlan,NUMOFVLAN,vlan);
 		}
 		return rv;
 	}
@@ -799,7 +986,7 @@ static int vlancreate_nocheck(int vlan)
 static int vlancreate(int vlan)
 {
 	if (vlan > 0 && vlan < NUMOFVLAN-1) { /*vlan NOVLAN (0xfff a.k.a. 4095) is reserved */
-		if (BAC_CHECK(validvlan,vlan))
+		if (bac_check(validvlan,vlan))
 			return EEXIST;
 		else 
 			return vlancreate_nocheck(vlan);
@@ -810,13 +997,13 @@ static int vlancreate(int vlan)
 static int vlanremove(int vlan)
 {
 	if (vlan >= 0 && vlan < NUMOFVLAN) {
-		if (BAC_CHECK(validvlan,vlan)) {
+		if (bac_check(validvlan,vlan)) {
 			register int i,used=0;
-			BA_FORALL(vlant[vlan].table,numports,used++,i);
+			ba_FORALL(vlant[vlan].table,numports,used++,i);
 			if (used)
 				return EADDRINUSE;
 			else {
-				BAC_CLR(validvlan,NUMOFVLAN,vlan);
+				bac_clr(validvlan,NUMOFVLAN,vlan);
 				free(vlant[vlan].table);
 				free(vlant[vlan].bctag);
 				free(vlant[vlan].bcuntag);
@@ -843,16 +1030,16 @@ static int vlanaddport(char *arg)
 		return EINVAL;
 	if (vlan <0 || vlan >= NUMOFVLAN-1 || port < 0 || port >= numports)
 		return EINVAL;
-	if (!BAC_CHECK(validvlan,vlan) || portv[port] == NULL)
+	if (!bac_check(validvlan,vlan) || portv[port] == NULL)
 		return ENXIO;
 	if (portv[port]->ep != NULL && portv[port]->vlanuntag != vlan) {
 		/* changing active port*/
-		BA_SET(vlant[vlan].bctag,port);
+		ba_set(vlant[vlan].bctag,port);
 #ifdef FSTP
 		fstaddport(vlan,port,1);
 #endif
 	}
-	BA_SET(vlant[vlan].table,port);
+	ba_set(vlant[vlan].table,port);
 	return 0;
 }
 
@@ -863,25 +1050,25 @@ static int vlandelport(char *arg)
 		return EINVAL;
 	if (vlan <0 || vlan >= NUMOFVLAN-1 || port < 0 || port >= numports)
 		return EINVAL;
-	if (!BAC_CHECK(validvlan,vlan) || portv[port] == NULL)
+	if (!bac_check(validvlan,vlan) || portv[port] == NULL)
 		return ENXIO;
 	if (portv[port]->vlanuntag == vlan)
 		return EADDRINUSE;
 	if (portv[port]->ep != NULL) {
 		/*changing active port*/
-		BA_CLR(vlant[vlan].bctag,port);
+		ba_clr(vlant[vlan].bctag,port);
 #ifdef FSTP
 		fstdelport(vlan,port);
 #endif
 	}
-	BA_CLR(vlant[vlan].table,port);
+	ba_clr(vlant[vlan].table,port);
 	hash_delete_port(port);
 	return 0;
 }
 
 #define STRSTATUS(PN,V) \
-	((BA_CHECK(vlant[(V)].notlearning,(PN))) ? "Discarding" : \
-	 (BA_CHECK(vlant[(V)].bctag,(PN)) || BA_CHECK(vlant[(V)].bcuntag,(PN))) ? \
+	((ba_check(vlant[(V)].notlearning,(PN))) ? "Discarding" : \
+	 (ba_check(vlant[(V)].bctag,(PN)) || ba_check(vlant[(V)].bcuntag,(PN))) ? \
 	 "Forwarding" : "Learning")
 
 static void vlanprintactive(int vlan,FILE *fd)
@@ -900,21 +1087,21 @@ static void vlanprintactive(int vlan,FILE *fd)
 				fsttab[vlan]->rootport, 
 				ntohl(*(u_int32_t *)(&(fsttab[vlan]->rootcost))),
 				qtime()-fsttab[vlan]->roottimestamp);
-		BA_FORALL(vlant[vlan].table,numports,
+		ba_FORALL(vlant[vlan].table,numports,
 				({ int tagged=portv[i]->vlanuntag != vlan;
 				 if (portv[i]->ep)
 				 printoutc(fd," -- Port %04d tagged=%d act=%d learn=%d forw=%d cost=%d role=%s",
 					 i, tagged, 1, !(NOTLEARNING(i,vlan)),
-					 (tagged)?(BA_CHECK(vlant[vlan].bctag,i) != 0):(BA_CHECK(vlant[vlan].bcuntag,i) != 0),
+					 (tagged)?(ba_check(vlant[vlan].bctag,i) != 0):(ba_check(vlant[vlan].bcuntag,i) != 0),
 					 portv[i]->cost,
 					 (fsttab[vlan]->rootport==i?"Root":
-						((BA_CHECK(fsttab[vlan]->backup,i)?"Alternate/Backup":"Designated")))
+						((ba_check(fsttab[vlan]->backup,i)?"Alternate/Backup":"Designated")))
 					 ); 0;
 				 }) ,i);
 #endif
 	} else {
 #endif
-		BA_FORALL(vlant[vlan].table,numports,
+		ba_FORALL(vlant[vlan].table,numports,
 				({ int tagged=portv[i]->vlanuntag != vlan;
 				 if (portv[i]->ep)
 				 printoutc(fd," -- Port %04d tagged=%d active=1 status=%s", i, tagged, 
@@ -931,14 +1118,14 @@ static int vlanprint(FILE *fd,char *arg)
 		register int vlan;
 		vlan=atoi(arg);
 		if (vlan >= 0 && vlan < NUMOFVLAN-1) {
-			if (BAC_CHECK(validvlan,vlan))
+			if (bac_check(validvlan,vlan))
 				vlanprintactive(vlan,fd);
 			else
 				return ENXIO;
 		} else
 			return EINVAL;
 	} else 
-		BAC_FORALLFUN(validvlan,NUMOFVLAN,vlanprintactive,fd);
+		bac_FORALLFUN(validvlan,NUMOFVLAN,vlanprintactive,fd);
 	return 0;
 }
 
@@ -946,7 +1133,7 @@ static void vlanprintelem(int vlan,FILE *fd)
 {
 	register int i;
 	printoutc(fd,"VLAN %04d",vlan);
-	BA_FORALL(vlant[vlan].table,numports,
+	ba_FORALL(vlant[vlan].table,numports,
 			printoutc(fd," -- Port %04d tagged=%d active=%d status=%s",
 				i, portv[i]->vlanuntag != vlan, portv[i]->ep != NULL, STRSTATUS(i,vlan)),i);
 }
@@ -957,14 +1144,14 @@ static int vlanprintall(FILE *fd,char *arg)
 		register int vlan;
 		vlan=atoi(arg);
 		if (vlan > 0 && vlan < NUMOFVLAN-1) {
-			if (BAC_CHECK(validvlan,vlan))
+			if (bac_check(validvlan,vlan))
 				vlanprintelem(vlan,fd);
 			else
 				return ENXIO;
 		} else
 			return EINVAL;
 	} else 
-		BAC_FORALLFUN(validvlan,NUMOFVLAN,vlanprintelem,fd);
+		bac_FORALLFUN(validvlan,NUMOFVLAN,vlanprintelem,fd);
 	return 0;
 }
 
@@ -1001,6 +1188,8 @@ static struct comlist cl[]={
 	{"port/create","N","create the port N (inactive|notallocatable)",portcreate,INTARG},
 	{"port/remove","N","remove the port N",portremove,INTARG},
 	{"port/allocatable","N 0/1","Is the port allocatable as unnamed? 1=Y 0=N",portallocatable,STRARG},
+	{"port/setuser","N user","access control: set user",portsetuser,STRARG},
+	{"port/setgroup","N user","access control: set group",portsetgroup,STRARG},
 	{"port/epclose","N ID","remove the endpoint port N/id ID",epclose,STRARG},
 #ifdef PORTCOUNTERS
 	{"port/resetcounter","[N]","reset the port (N) counters",portresetcounters,STRARG},
@@ -1024,7 +1213,7 @@ void port_init(int initnumports)
 	}
 	portv=calloc(numports,sizeof(struct port *));
 	/* vlan_init */
-	validvlan=BAC_ALLOC(NUMOFVLAN);
+	validvlan=bac_alloc(NUMOFVLAN);
 	if (portv==NULL || validvlan == NULL) {
 		printlog(LOG_ERR,"ALLOC port data structures");
 		exit(1);
